__author__ = "Jean-Pierre LESUEUR (@DarkCoderSc)"
__maintainer__ = "Jean-Pierre LESUEUR"
__email__ = "jplesueur@phrozen.io"
__copyright__ = "Copyright 2024, Phrozen"
__license__ = "Apache License 2.0"

from .about import AboutDialog
from .connecting import ConnectingDialog
from .options import OptionsDialog
from .screen_selection import ScreenSelectionDialog
from .server_certificate import ServerCertificateDialog

__all__ = [
    'AboutDialog',
    'ConnectingDialog',
    'ScreenSelectionDialog',
    'ServerCertificateDialog',
    'OptionsDialog',
]



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.
"""

from typing import Optional, Union

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (QCheckBox, QDialog, QHBoxLayout, QLabel,
                             QMainWindow, QPushButton, QVBoxLayout)

import arcane_viewer.ui.utilities as utilities


class ServerCertificateDialog(utilities.QCenteredDialog):
    def __init__(self, parent: Optional[Union[QDialog, QMainWindow]], fingerprint: str) -> None:
        super().__init__(parent)

        self.setWindowTitle("Unknown Server Certificate")

        self.setWindowFlags(
            Qt.WindowType.Dialog |
            Qt.WindowType.MSWindowsFixedSizeDialogHint
        )

        core_layout = QVBoxLayout()
        self.setLayout(core_layout)

        # Certificate Information
        certificate_info_label = QLabel("Please ensure the following fingerprint (SHA-1) matches the expected server "
                                        "fingerprint.")

        certificate_info_label.setObjectName("alert-warning")
        certificate_info_label.setWordWrap(True)

        core_layout.addWidget(certificate_info_label)

        core_layout.addSpacing(16)

        self.fingerprint = fingerprint

        core_layout.addLayout(self.setup_fingerprint_layout(0, 10))
        core_layout.addLayout(self.setup_fingerprint_layout(10, 20))
        core_layout.addLayout(self.setup_fingerprint_layout(20, 30))
        core_layout.addLayout(self.setup_fingerprint_layout(30, 40))

        core_layout.addSpacing(8)

        # Trust Certificate Option for next sessions
        self.trust_certificate_checkbox = QCheckBox("Add this certificate to the trusted certificates store")
        core_layout.addWidget(self.trust_certificate_checkbox)

        core_layout.addSpacing(8)

        # Action Buttons
        action_buttons_layout = QHBoxLayout()
        core_layout.addLayout(action_buttons_layout)

        self.reject_button = QPushButton('Reject')
        self.reject_button.clicked.connect(lambda: self.reject())
        action_buttons_layout.addWidget(self.reject_button)

        self.accept_button = QPushButton('Accept')
        self.accept_button.clicked.connect(lambda: self.accept())
        action_buttons_layout.addWidget(self.accept_button)

        self.reject_button.setDefault(True)
        self.reject_button.setFocus()

        self.adjust_size()

    def setup_fingerprint_layout(self, start: int, end: int) -> QHBoxLayout:
        layout = QHBoxLayout()
        line = [self.fingerprint[i:i + 2] for i in range(start, end, 2)]

        for index, item in enumerate(line):
            label = QLabel(item)
            label.setFont(utilities.MONOSPACE_FONTS)
            label.setObjectName("fingerprint")
            layout.addWidget(label)

            if index < len(line) - 1:
                label = QLabel(" : ")
                label.setAlignment(Qt.AlignmentFlag.AlignCenter)
                layout.addWidget(label)

        return layout

    def adjust_size(self) -> None:
        self.setFixedSize(
            self.sizeHint().width(),
            self.sizeHint().height()
        )



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io

    Todo:
        - Find a workaround for the Wayland issue for positioning the window.
        - Offer connection profile (Add/Edit/Delete)
        - Implement Clipboard sharing between client and server.
        - Offer possibility to change the default theme with something retro and cool.
"""

import logging
import sys
import threading

from PyQt6.QtGui import QColor, QIcon, QPalette
from PyQt6.QtWidgets import QApplication

import arcane_viewer.arcane as arcane
import arcane_viewer.ui.forms as arcane_forms


def main() -> None:
    logging.basicConfig(
                        level=logging.DEBUG,
                        format="%(asctime)s - %(name)s[%(thread)d] - %(levelname)s - %(message)s"
    )

    app = QApplication(sys.argv)

    app.setWindowIcon(QIcon(arcane.APP_ICON))

    app.setStyle('Fusion')

    # Create and apply a dark palette
    dark_palette = QPalette()
    dark_palette.setColor(QPalette.ColorRole.Window, QColor(53, 53, 53))
    dark_palette.setColor(QPalette.ColorRole.WindowText, QColor(255, 255, 255))
    dark_palette.setColor(QPalette.ColorRole.Base, QColor(42, 42, 42))
    dark_palette.setColor(QPalette.ColorRole.AlternateBase, QColor(66, 66, 66))
    dark_palette.setColor(QPalette.ColorRole.ToolTipBase, QColor(255, 255, 255))
    dark_palette.setColor(QPalette.ColorRole.ToolTipText, QColor(255, 255, 255))
    dark_palette.setColor(QPalette.ColorRole.Text, QColor(255, 255, 255))
    dark_palette.setColor(QPalette.ColorRole.Button, QColor(53, 53, 53))
    dark_palette.setColor(QPalette.ColorRole.ButtonText, QColor(255, 255, 255))
    dark_palette.setColor(QPalette.ColorRole.BrightText, QColor(255, 0, 0))
    dark_palette.setColor(QPalette.ColorRole.Link, QColor(42, 130, 218))
    dark_palette.setColor(QPalette.ColorRole.Highlight, QColor(42, 130, 218))
    dark_palette.setColor(QPalette.ColorRole.HighlightedText, QColor(0, 0, 0))

    app.setPalette(dark_palette)

    # Apply custom stylesheet
    app.setStyleSheet("""
        QGroupBox::title {
            background-color: rgb(53, 53, 53);
            padding: 4px 8px 4px 8px;
            border-radius: 4px;
            left: 8px;
            top: 8px;
        }

        QLabel#fingerprint {
            background-color: rgb(42, 42, 42);
            padding: 4px;
            border-radius: 4px;
        }

        QLabel#alert-warning {
            background-color: #d0b650;
            color: #564917;
            padding: 4px;
            border-radius: 4px;
        }

        QHeaderView::section {
            height: 26px;
        }

        QTabBar::tab {
            height: 26px;
        }

        QPushButton#danger {
            background-color: darkred;
        }

        QPushButton:disabled {
            background-color: #444444;
            color: #777777;
        }
        """)

    logging.debug(f"Main thread ID: {threading.get_ident()}")

    # Create and show the connect window
    connect_window = arcane_forms.ConnectWindow()
    connect_window.show()

    sys.exit(app.exec())


if __name__ == '__main__':
    main()



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.
"""

import sys
from typing import Optional, Union

from PyQt6.QtCore import QT_VERSION_STR, QSize, Qt
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import (QDialog, QHBoxLayout, QLabel, QMainWindow,
                             QPushButton, QVBoxLayout)

import arcane_viewer.arcane as arcane
import arcane_viewer.ui.utilities as utilities


class AboutDialog(utilities.QCenteredDialog):
    def __init__(self, parent: Optional[Union[QDialog, QMainWindow]] = None) -> None:
        super().__init__(parent)

        self.setWindowTitle(f"About {arcane.APP_DISPLAY_NAME}")

        self.setWindowFlags(
            Qt.WindowType.Dialog |
            Qt.WindowType.CustomizeWindowHint |
            Qt.WindowType.WindowCloseButtonHint |
            Qt.WindowType.MSWindowsFixedSizeDialogHint
        )

        # Setup Main Layout (Core Layout)
        core_layout = QVBoxLayout()
        self.setLayout(core_layout)

        # Avatar
        avatar = QLabel(self)
        avatar.setPixmap(QIcon(arcane.APP_ICON).pixmap(QSize(96, 96)))
        avatar.setAlignment(Qt.AlignmentFlag.AlignCenter)
        core_layout.addWidget(avatar)

        # About Author / Project
        text = QLabel(f"<h1>{arcane.APP_DISPLAY_NAME}</h1>"
                      f"<p><b>Protocol Version:</b> {arcane.PROTOCOL_VERSION}</p>"
                      f"<p><b>Python:</b> {sys.version.split()[0]} / <b>PyQt6: </b>{QT_VERSION_STR}</p>"
                      "<p>🇫🇷 Jean-Pierre LESUEUR (<a href=\"https://twitter.com/darkcodersc\">@DarkCoderSc</a>)</p>"
                      "<p><b>License:</b> Apache License 2.0</p>"
                      "<p><a href=\"https://github.com/PhrozenIO\">www.github.com/PhrozenIO</a></p>"
                      "<p>© 2024 - <a href=\"https://www.phrozen.io\">www.phrozen.io</a></p>")
        text.setWordWrap(True)
        text.setAlignment(Qt.AlignmentFlag.AlignCenter)
        core_layout.addWidget(text)

        # Give breath to the layout (before action buttons)
        core_layout.addSpacing(8)

        # Action Button(s)
        action_buttons_layout = QHBoxLayout()
        core_layout.addLayout(action_buttons_layout)

        action_buttons_layout.addStretch(1)

        close_button = QPushButton("Close")
        close_button.clicked.connect(self.close)
        action_buttons_layout.addWidget(close_button)

        action_buttons_layout.addStretch(1)

        self.adjust_size()

    def adjust_size(self) -> None:
        self.setFixedSize(380, self.sizeHint().height())



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.
"""

from PyQt6.QtCore import QSize


class Screen:
    """ Screen class to store screen information """
    def __init__(self, screen_information_json: dict) -> None:
        if not all(k in screen_information_json for k in (
                "Id",
                "Name",
                "Width",
                "Height",
                "X",
                "Y",
                "Primary",
        )):
            raise Exception("Invalid screen information received")

        self.id = screen_information_json["Id"]
        self.name = screen_information_json["Name"]
        self.width = screen_information_json["Width"]
        self.height = screen_information_json["Height"]
        self.x = screen_information_json["X"]
        self.y = screen_information_json["Y"]
        self.primary = screen_information_json["Primary"]

    def get_display_name(self) -> str:
        return "#{} - {} ({}x{})".format(
            self.id,
            self.name,
            self.width,
            self.height
        )

    def size(self) -> QSize:
        return QSize(self.width, self.height)



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.
"""

import logging
import traceback
from abc import abstractmethod
from typing import Optional

from PyQt6.QtCore import QMutex, QThread, pyqtSignal, pyqtSlot

import arcane_viewer.arcane as arcane

logger = logging.getLogger(__name__)


class ClientBaseThread(QThread):
    """ Base class for all client threads """
    thread_finished = pyqtSignal(bool)

    """`Destruction is a form of creation. So the fact they burn the money is ironic. They just want to see what happens
     when they tear the world apart. They want to change things.`, Donnie Darko"""
    def __init__(self, session: arcane.Session, worker_kind: arcane.WorkerKind) -> None:
        super().__init__()

        self._running = True
        self._connected = False
        self._mutex = QMutex()

        self.session = session
        self.client: Optional[arcane.Client] = None
        self.worker_kind = worker_kind

    def run(self) -> None:
        on_error = False
        try:
            self.client = self.session.claim_client(self.worker_kind)

            self._connected = True

            # Implement me :-)
            self.client_execute()

            on_error = False

            logger.debug(f"`{self.__class__.__name__}` Thread gracefully ended.")
        except Exception as e:
            if self._running:
                logger.error(f"Thread `{self.__class__.__name__}` encountered an error: `{e}`")
                traceback.print_exc()
                on_error = True
        finally:
            self.stop()

            self.thread_finished.emit(on_error)

    @abstractmethod
    def client_execute(self) -> None:
        pass

    @pyqtSlot()
    def stop(self) -> None:
        self._mutex.lock()
        try:
            if self.client is not None:
                self.client.close()

            self._running = False
        finally:
            self._mutex.unlock()



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.
"""

import json
import logging
from typing import Optional

from PyQt6.QtCore import QSettings

import arcane_viewer.arcane as arcane

logger = logging.getLogger(__name__)


class Session:
    """ Session class to handle remote session """
    def __init__(self, server_address: str, server_port: int, password: str) -> None:
        self.server_address = server_address
        self.server_port = server_port
        self.__password = password

        self.presentation = False

        self.session_id: Optional[str] = None
        self.display_name: Optional[str] = None
        self.server_fingerprint: Optional[str] = None

        # Load settings (options)
        settings = QSettings(arcane.APP_ORGANIZATION_NAME, arcane.APP_NAME)

        # Remote Desktop Options
        self.clipboard_mode = settings.value(arcane.SETTINGS_KEY_CLIPBOARD_MODE, arcane.ClipboardMode.Both)

        # Remote Desktop Capture Options
        self.option_image_quality = settings.value(arcane.SETTINGS_KEY_IMAGE_QUALITY, 80)
        self.option_packet_size = settings.value(arcane.SETTINGS_KEY_PACKET_SIZE, arcane.PacketSize.Size4096)
        self.option_block_size = settings.value(arcane.SETTINGS_KEY_BLOCK_SIZE, arcane.BlockSize.Size64)

        self.request_session()

    def claim_client(self, worker_kind: Optional[arcane.WorkerKind] = None) -> arcane.Client:
        """ Establish a new TLS connection to the remote server and authenticate. Optionally we can specify a worker
        to be attached to the current session """
        client = arcane.Client(self.server_address, self.server_port, self.__password)

        # If a session is already established and a worker kind is provided, we attach to current session a new worker
        if worker_kind is not None:
            if self.session_id is None:
                raise arcane.ArcaneProtocolException(arcane.ArcaneProtocolError.MissingSession)

            # If the server fingerprint has changed after session creation, we may be facing a MITM attack,
            # Abort connection
            if self.server_fingerprint != client.server_fingerprint:
                raise arcane.ArcaneProtocolException(arcane.ArcaneProtocolError.ServerFingerprintTampered)

            client.write_line("AttachToSession")

            client.write_line(self.session_id)

            response = client.read_line()
            if response != "ResourceFound":
                raise arcane.ArcaneProtocolException(arcane.ArcaneProtocolError.ResourceNotFound)

            client.write_line(worker_kind.name)

        return client

    def request_session(self) -> None:
        """ Request a new session to the remote server """
        client = self.claim_client()
        try:
            self.server_fingerprint = client.server_fingerprint

            client.write_line("RequestSession")

            session_information = client.read_json()
            if session_information is None:
                raise arcane.ArcaneProtocolException(arcane.ArcaneProtocolError.InvalidStructureData)

            logger.debug("@Session information:")
            logger.debug(json.dumps(session_information, indent=4))

            if not all(k in session_information for k in (
                    "SessionId",
                    "Version",
                    "ViewOnly",
                    "Clipboard",
                    "Username",
                    "MachineName",
                    "WindowsVersion",
            )):
                raise arcane.ArcaneProtocolException(arcane.ArcaneProtocolError.InvalidStructureData)

            # Protocol Version Check
            if session_information["Version"] != arcane.PROTOCOL_VERSION:
                logger.error(f"Incompatible server version, client version: `{arcane.PROTOCOL_VERSION}` != "
                             f"server version: `{session_information['Version']}`")

                raise arcane.ArcaneProtocolException(
                    arcane.ArcaneProtocolError.UnsupportedVersion
                )

            # Assign session information
            self.session_id = session_information["SessionId"]

            self.display_name = "{}@{}".format(
                session_information["Username"],
                session_information["MachineName"],
            )

            # Handle Server-Clipboard Mode and solve possible clash with client-clipboard mode
            # By default, if for any reason, server clipboard mode is not recognized, we consider it as disabled
            server_clipboard_mode = arcane.ClipboardMode.Disabled
            try:
                server_clipboard_mode = arcane.ClipboardMode(session_information["Clipboard"])
            except ValueError:
                pass

            logger.info("Server clipboard mode: `{}`".format(server_clipboard_mode.name))

            if session_information["ViewOnly"]:
                logger.warning("Presentation mode enforced by remote server, no input / output "
                               "(Mouse, Keyboard, Clipboard) will be accepted")

                # In view only (presentation) mode, whatever the client clipboard mode is, we disable it
                self.clipboard_mode = arcane.ClipboardMode.Disabled
                self.presentation = True
            else:
                # If server clipboard mode is disabled, whatever the client clipboard mode is, we disable it
                if (server_clipboard_mode == arcane.ClipboardMode.Disabled and
                        self.clipboard_mode != arcane.ClipboardMode.Disabled):
                    self.clipboard_mode = arcane.ClipboardMode.Disabled
                    logger.warning("Server clipboard mode is disabled, reflecting to client clipboard mode")
                # If server clipboard mode is send-only, we set client clipboard mode to receive-only only if client
                # clipboard mode is set to both.
                elif (server_clipboard_mode == arcane.ClipboardMode.Send and
                      self.clipboard_mode == arcane.ClipboardMode.Both):
                    self.clipboard_mode = arcane.ClipboardMode.Receive
                    logger.warning("Server clipboard mode is send-only, client clipboard mode set to receive-only")
                # If server clipboard mode is receive-only, we set client clipboard mode to send-only only if client
                # clipboard mode is set to both.
                elif (server_clipboard_mode == arcane.ClipboardMode.Receive and
                      self.clipboard_mode == arcane.ClipboardMode.Both):
                    self.clipboard_mode = arcane.ClipboardMode.Send
                    logger.warning("Server clipboard mode is receive-only, client clipboard mode set to send-only")
                # If there is a clash between server and client clipboard mode, we disable it
                elif (
                        (server_clipboard_mode == arcane.ClipboardMode.Send and
                         self.clipboard_mode == arcane.ClipboardMode.Send) or
                        (server_clipboard_mode == arcane.ClipboardMode.Receive and
                         self.clipboard_mode == arcane.ClipboardMode.Receive)
                ):
                    self.clipboard_mode = arcane.ClipboardMode.Disabled
                    logger.warning("Server clipboard mode and client clash, clipboard is then disabled")

            # Finally
            logger.info("Session established with `{}` on `{}`".format(
                self.display_name,
                session_information["WindowsVersion"],
            ))
        finally:
            if client is not None:
                client.close()



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.
"""

from enum import Enum, auto


class ArcaneProtocolError(Enum):
    AuthenticationFailed = auto()
    ResourceNotFound = auto()
    InvalidStructureData = auto()
    UnsupportedVersion = auto()
    MissingServerCertificate = auto()
    MissingSession = auto()
    ServerFingerprintTampered = auto()


class ArcaneProtocolException(Exception):
    def __init__(self, reason: ArcaneProtocolError) -> None:
        self.reason = reason

        error_messages = {
            ArcaneProtocolError.AuthenticationFailed: "Authentication failed",
            ArcaneProtocolError.ResourceNotFound: "A resource was not found",
            ArcaneProtocolError.InvalidStructureData: "Invalid structure data, whether corrupted or missing properties",
            ArcaneProtocolError.UnsupportedVersion: "Unsupported protocol version: protocol versions between client and"
                                                    "server do not match",
            ArcaneProtocolError.MissingServerCertificate: "Server certificate is missing",
            ArcaneProtocolError.MissingSession: "Session is missing, you must request a session first",
            ArcaneProtocolError.ServerFingerprintTampered: "Server fingerprint has changed, confidentiality is possibly"
                                                           " compromised",
        }

        super().__init__(error_messages.get(reason, "Unknown error"))



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.
"""

from typing import Optional

from PyQt6.QtGui import QFont, QShowEvent
from PyQt6.QtWidgets import QDialog, QMainWindow, QWidget

MONOSPACE_FONTS = QFont("Consolas, 'Courier New', Monaco, 'DejaVu Sans Mono', 'Liberation Mono', monospace")


class CenteredWindowMixin(QWidget):
    """ Mixin to center a widget on the screen or on another widget """
    def center_on_owner(self) -> None:
        parent = self.parent()
        screen = self.screen()

        if parent is not None and isinstance(parent, QWidget):
            owner_geometry = parent.frameGeometry()
        elif screen is not None:
            owner_geometry = screen.availableGeometry()
        else:
            return

        subform_geometry = self.frameGeometry()

        center_point = owner_geometry.center()

        subform_geometry.moveCenter(center_point)
        self.move(subform_geometry.topLeft())


class QCenteredDialog(QDialog, CenteredWindowMixin):
    def showEvent(self, event: Optional[QShowEvent]) -> None:
        self.center_on_owner()


class QCenteredMainWindow(QMainWindow, CenteredWindowMixin):
    def showEvent(self, event: Optional[QShowEvent]) -> None:
        self.center_on_owner()



__author__ = "Jean-Pierre LESUEUR (@DarkCoderSc)"
__maintainer__ = "Jean-Pierre LESUEUR"
__email__ = "jplesueur@phrozen.io"
__copyright__ = "Copyright 2024, Phrozen"
__license__ = "Apache License 2.0"

from .tangeant_universe import TangentUniverse

__all__ = [
    'TangentUniverse',
]



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.

    Todo:
        - (0001) : Implement remote screen resolution update, if it does, we will need to ensure it push an update of
                   selected screen object and reflect the changes where necessary.
"""

import logging
from typing import List, Optional, Union

from PyQt6.QtCore import QRect, QSize, Qt, pyqtSlot
from PyQt6.QtGui import (QCloseEvent, QImage, QPainter, QPixmap, QResizeEvent,
                         QScreen, QShowEvent, QTransform)
from PyQt6.QtWidgets import (QApplication, QDialog, QGraphicsPixmapItem,
                             QMainWindow, QMessageBox)

import arcane_viewer.arcane as arcane
import arcane_viewer.arcane.threads as arcane_threads
import arcane_viewer.ui.custom_widgets as arcane_widgets
import arcane_viewer.ui.dialogs as arcane_dialogs

logger = logging.getLogger(__name__)


class DesktopWindow(QMainWindow):
    def __init__(self, connect_window: Union[QDialog, QMainWindow], session: arcane.Session) -> None:
        super().__init__()

        self.scene_pixmap: Optional[QGraphicsPixmapItem] = None
        self.v_desktop: Optional[QPixmap] = None

        self.desktop_thread: Optional[arcane_threads.VirtualDesktopThread] = None
        self.events_thread: Optional[arcane_threads.EventsThread] = None

        self.session = session

        # Instead of using QWidget parent property, we will use a custom attribute to store the "parent" window, this
        # will prevent icon to disappear on Windows taskbar when a parent is set to a Window but parent is hidden.
        self.connect_window = connect_window

        # Set Window Properties, Layout, Title, Icon and Size
        self.setWindowTitle("🖥 {} ({}) :: {} {}".format(
            arcane.APP_DISPLAY_NAME,
            session.server_address,
            session.display_name,
            "- View Only" if session.presentation else ""
        ))
        self.resize(QSize(640, 360))
        self.setMouseTracking(True)

        self.setContentsMargins(0, 0, 0, 0)

        self.tangent_universe = arcane_widgets.TangentUniverse()
        self.setCentralWidget(self.tangent_universe)

        """
        # FPS Counter (Debugging)
        self.FPS_counter = 0
        self.FPS_Elapsed = time.time()
        """

        self.start_desktop_thread()

    def thread_finished(self, on_error: bool) -> None:
        """ Handle the thread finished event """
        if on_error:
            QMessageBox.critical(self, "Error", "Something went wrong, check console output for more information.")

        self.close()

    def start_desktop_thread(self) -> None:
        """ Start the desktop thread to handle remote desktop streaming """
        self.desktop_thread = arcane_threads.VirtualDesktopThread(self.session)
        self.desktop_thread.chunk_received.connect(self.update_scene)
        self.desktop_thread.open_cellar_door.connect(self.open_cellar_door)
        self.desktop_thread.thread_finished.connect(self.thread_finished)
        self.desktop_thread.request_screen_selection.connect(self.display_screen_selection_dialog)
        self.desktop_thread.start()

    def start_events_thread(self, screen: arcane.Screen) -> None:
        """ Start the events thread to handle remote desktop events """
        self.events_thread = arcane_threads.EventsThread(self.session)
        self.events_thread.thread_finished.connect(self.thread_finished)
        self.events_thread.start()

        # Assign our events thread to the Tangent Universe
        self.tangent_universe.set_event_thread(self.events_thread)
        self.tangent_universe.set_screen(screen)

    def close_cellar_door(self) -> None:
        """ Collapse Tangent Universe to Main Branch, We were able to save the world before 28:06:42:12 """
        if self.desktop_thread is not None:
            if self.desktop_thread.isRunning():
                self.desktop_thread.stop()
                self.desktop_thread.wait()

        if self.events_thread is not None:
            if self.events_thread.isRunning():
                self.events_thread.stop()
                self.events_thread.wait()

    def showEvent(self, event: Optional[QShowEvent]) -> None:
        super().showEvent(event)

        if self.connect_window is not None:
            self.connect_window.hide()

    def closeEvent(self, event: Optional[QCloseEvent]) -> None:
        """ Overridden close method to handle the cleanup
            `I Hope That When The World Comes To An End, I Can Breathe A Sigh Of Relief Because There Will Be So Much To
             Look Forward To.`"""

        self.close_cellar_door()

        if event is not None:
            event.accept()

        if self.connect_window is not None:
            self.connect_window.show()

    def open_cellar_door(self, screen: arcane.Screen) -> None:
        """ Initialize the virtual desktop (Tangent Universe) """
        self.v_desktop = QPixmap(screen.size())
        self.v_desktop.fill(Qt.GlobalColor.black)

        self.scene_pixmap = QGraphicsPixmapItem(self.v_desktop)
        self.tangent_universe.desktop_scene.addItem(self.scene_pixmap)

        # Initialize the size of virtual desktop window regarding our current monitor screen size
        local_screen: Optional[QScreen] = None

        for local_screen_candidate in QApplication.screens():
            if local_screen_candidate.geometry().intersects(self.geometry()):
                local_screen = local_screen_candidate

                break

        if local_screen is None:
            local_screen = QApplication.primaryScreen()

        if local_screen is None:
            logger.error("Unable to find a valid screen to initialize the virtual desktop window.")

            self.close()

            return

        local_screen_size = local_screen.size()
        remote_screen_width = int(screen.width / local_screen.devicePixelRatio())
        remote_screen_height = int(screen.height / local_screen.devicePixelRatio())

        # Calculate the new width and height of the virtual desktop window
        if local_screen_size.width() <= remote_screen_width or local_screen_size.height() <= remote_screen_height:
            adjust_vertically = local_screen_size.width() > local_screen_size.height()

            if adjust_vertically:
                new_width = round((local_screen_size.width() * arcane.VD_WINDOW_ADJUST_RATIO) / 100)
                resized_ratio = round(new_width * 100 / remote_screen_width)
                new_height = round((remote_screen_height * resized_ratio) / 100)
            else:
                new_height = round((local_screen_size.height() * arcane.VD_WINDOW_ADJUST_RATIO) / 100)
                resized_ratio = round(new_height * 100 / remote_screen_height)
                new_width = round((remote_screen_width * resized_ratio) / 100)
        else:
            new_width = remote_screen_width
            new_height = remote_screen_height

        # Resize and place the window
        self.setGeometry(
            local_screen.geometry().left() + (local_screen_size.width() - new_width) // 2,
            local_screen.geometry().top() + (local_screen_size.height() - new_height) // 2,
            new_width,
            new_height
        )

        # We can now start our events thread
        # TODO: 0001
        self.start_events_thread(screen)

    def fit_scene(self) -> None:
        """ Fit the scene (Hacky Technique) to the view """
        if (
                self.tangent_universe is None or
                self.scene_pixmap is None or
                self.v_desktop is None
        ):
            return

        # Instead of bellow code:
        #   `self.view.fitInView(self.scene_pixmap, Qt.AspectRatioMode.IgnoreAspectRatio)`
        # We will calculate the scale factor manually to avoid the aspect ratio issue and fitting correctly the view to
        # our virtual desktop host window.
        view_rect = self.tangent_universe.frameRect()

        scale_x = view_rect.width() / self.v_desktop.width()
        scale_y = view_rect.height() / self.v_desktop.height()

        transform = QTransform()
        transform.scale(scale_x, scale_y)
        self.tangent_universe.setTransform(transform, False)

    def update_scene(self, chunk: QImage, x: int, y: int) -> None:
        """ Update the virtual desktop with the received chunk """
        if self.v_desktop is None or self.scene_pixmap is None:
            return

        if chunk is None or not isinstance(chunk, QImage):
            return

        # Update the virtual desktop with the received chunk (Tangent Universe)
        rect = QRect(x, y, chunk.width(), chunk.height())

        painter = QPainter(self.v_desktop)
        painter.drawImage(rect, chunk)
        painter.end()

        # Update the scene with the updated virtual desktop
        self.scene_pixmap.setPixmap(self.v_desktop)

        self.fit_scene()

        """
        # FPS Counter (Debugging)
        self.FPS_counter += 1
        elapsed = time.time() - self.FPS_Elapsed
        if elapsed >= 1.0:
            logger.debug("FPS: {}".format(self.FPS_counter))
            self.FPS_counter = 0
            self.FPS_Elapsed = time.time()
        """

    def resizeEvent(self, event: Optional[QResizeEvent]) -> None:
        """ Overridden resizeEvent method to fit the scene to the view """
        self.fit_scene()

    def screen_selection_rejected(self) -> None:
        self.close()

    @pyqtSlot(list)
    def display_screen_selection_dialog(self, screens: List[arcane.Screen]) -> None:
        """ Display screen selection dialog """
        screen_selection_dialog = arcane_dialogs.ScreenSelectionDialog(self, screens)

        screen_selection_dialog.accepted.connect(lambda: self.desktop_thread.on_screen_selection_dialog_closed(
            screen_selection_dialog.get_selected_screen()
        ) if self.desktop_thread is not None else None)

        screen_selection_dialog.rejected.connect(self.screen_selection_rejected)
        screen_selection_dialog.exec()



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.
"""

import json
import os.path
import socket
from typing import Optional

from PyQt6.QtCore import QSettings, QSize, Qt, pyqtSlot
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import (QDialog, QHBoxLayout, QLabel, QLineEdit,
                             QMessageBox, QPushButton, QSpinBox, QVBoxLayout,
                             QWidget)

import arcane_viewer.arcane as arcane
import arcane_viewer.arcane.threads as arcane_threads
import arcane_viewer.ui.dialogs as arcane_dialogs
import arcane_viewer.ui.forms as arcane_forms
import arcane_viewer.ui.utilities as utilities


class ConnectWindow(utilities.QCenteredMainWindow):
    """ Connect Window to establish a connection to the server """

    def __init__(self) -> None:
        super().__init__()

        self.__connect_thread: Optional[arcane_threads.ConnectThread] = None
        self.__connecting_dialog: Optional[arcane_dialogs.ConnectingDialog] = None
        self.desktop_window: Optional[arcane_forms.DesktopWindow] = None
        self.session: Optional[arcane.Session] = None

        self.setWindowTitle(f"{arcane.APP_DISPLAY_NAME} :: Connect")

        self.setWindowFlags(
            Qt.WindowType.CustomizeWindowHint |
            Qt.WindowType.WindowCloseButtonHint |
            Qt.WindowType.MSWindowsFixedSizeDialogHint
        )

        # Create central widget
        self.central_widget = QWidget()
        self.setCentralWidget(self.central_widget)

        # Setup Main Layout (Core Layout)
        core_layout = QVBoxLayout()
        self.central_widget.setLayout(core_layout)

        # Setup Form Layout (Containing both Logo and Inputs)
        form_layout = QHBoxLayout()
        core_layout.addLayout(form_layout)

        # Logo
        icon = QLabel(self)
        icon.setPixmap(QIcon(arcane.APP_ICON).pixmap(QSize(96, 96)))
        icon.setAlignment(Qt.AlignmentFlag.AlignCenter)
        form_layout.addWidget(icon)

        # Form Inputs
        form_input_layout = QVBoxLayout()
        form_layout.addLayout(form_input_layout)

        # Server Address Form Input
        self.server_address_label = QLabel("Server Address / Port:")
        form_input_layout.addWidget(self.server_address_label)

        server_address_layout = QHBoxLayout()

        self.server_address_input = QLineEdit()
        self.server_address_input.setText("127.0.0.1")
        server_address_layout.addWidget(self.server_address_input)

        separator_label = QLabel(":")
        server_address_layout.addWidget(separator_label)

        self.server_port_input = QSpinBox()
        self.server_port_input.setMinimum(0)
        self.server_port_input.setMaximum(65535)
        self.server_port_input.setValue(2801)
        server_address_layout.addWidget(self.server_port_input)

        server_address_layout.setStretch(0, 5)
        server_address_layout.setStretch(1, 0)
        server_address_layout.setStretch(2, 2)

        form_input_layout.addLayout(server_address_layout)

        form_input_layout.addSpacing(4)

        # Password Form Input
        self.password_label = QLabel("Password:")
        form_input_layout.addWidget(self.password_label)

        self.password_input = QLineEdit()
        self.password_input.setEchoMode(QLineEdit.EchoMode.Password)

        form_input_layout.addWidget(self.password_input)

        # Give Breath to Action Buttons
        form_input_layout.addSpacing(6)

        # Action Buttons
        self.about_button = QPushButton("About")
        self.about_button.clicked.connect(self.show_about_dialog)

        self.options_button = QPushButton("Options")
        self.options_button.clicked.connect(lambda: arcane_dialogs.OptionsDialog(self).exec())

        self.connect_button = QPushButton("Connect")
        self.connect_button.clicked.connect(self.submit_form)
        self.connect_button.setDefault(True)

        action_layout = QHBoxLayout()
        action_layout.addWidget(self.about_button)
        action_layout.addWidget(self.options_button)
        action_layout.addWidget(self.connect_button)

        core_layout.addLayout(action_layout)

        # Read Default Settings: Mostly for development purposes, until custom profiles are implemented
        self.read_default()

        self.adjust_size()

    def read_default(self) -> None:
        """ Read default settings from the default.json file """
        if not os.path.isfile(arcane.DEFAULT_JSON):
            return

        with open(arcane.DEFAULT_JSON, "r") as f:
            try:
                data = json.load(f)
            except json.JSONDecodeError:
                return

            if "use" in data and data["use"] is False:
                return

            if "server_address" in data:
                self.server_address_input.setText(data["server_address"])

            if "server_port" in data:
                self.server_port_input.setValue(data["server_port"])

            if "server_password" in data:
                self.password_input.setText(data["server_password"])

    def submit_form(self) -> None:
        """ Validate the form and submit it """
        try:
            # Check if the ip/hostname is valid
            hostname = self.server_address_input.text()
            try:
                if socket.gethostbyname(hostname) == hostname:
                    pass
            except socket.gaierror:
                self.server_address_input.setFocus()
                raise Exception("Invalid hostname or IP address.")

            # Check password input
            if len(self.password_input.text().strip()) == 0:
                self.password_input.setFocus()
                raise Exception("Password field cannot be empty.")

            # Attempt connection
            self.__connect_thread = arcane_threads.ConnectThread(
                self.server_address_input.text(),
                self.server_port_input.value(),
                self.password_input.text(),
            )

            self.__connect_thread.thread_started.connect(self.connect_thread_started)
            self.__connect_thread.session_error.connect(self.session_error)
            self.__connect_thread.thread_finished.connect(self.connect_thread_finished)
            self.__connect_thread.start()
        except Exception as e:
            QMessageBox.critical(self, "Form Error", str(e))

    def adjust_size(self) -> None:
        self.setFixedSize(350, self.sizeHint().height())

    def show_about_dialog(self) -> None:
        about_window = arcane_dialogs.AboutDialog(self)
        about_window.exec()

    @pyqtSlot(str)
    def session_error(self, error_message: str) -> None:
        QMessageBox.critical(self, "Error", error_message)

    @pyqtSlot()
    def connect_thread_started(self) -> None:
        self.__connecting_dialog = arcane_dialogs.ConnectingDialog(self)
        self.__connecting_dialog.exec()

    @pyqtSlot(object)
    def connect_thread_finished(self, session: Optional[arcane.Session] = None) -> None:
        # Close the connecting form if it is still open
        if self.__connecting_dialog is not None and self.__connecting_dialog.isVisible():
            self.__connecting_dialog.close()

        if session is None or session.server_fingerprint is None:
            return

        self.session = session

        # Show Server Certificate Dialog if certificate is not (yet) trusted
        settings = QSettings(arcane.APP_ORGANIZATION_NAME, arcane.APP_NAME)
        trusted_certificates = settings.value(arcane.SETTINGS_KEY_TRUSTED_CERTIFICATES, [])

        if session.server_fingerprint not in trusted_certificates:
            server_certificate_dialog = arcane_dialogs.ServerCertificateDialog(self, session.server_fingerprint)
            signal = server_certificate_dialog.exec()
            if signal == QDialog.DialogCode.Rejected:
                return

            if server_certificate_dialog.trust_certificate_checkbox.isChecked():
                trusted_certificates.append(session.server_fingerprint)

                # Save trusted certificates to settings store
                settings.setValue(arcane.SETTINGS_KEY_TRUSTED_CERTIFICATES, trusted_certificates)

                # Save extra information about the certificate (Default)
                settings.setValue(f"{arcane.SETTINGS_KEY_TRUSTED_CERTIFICATES}.{session.server_fingerprint}", {
                    "display_name": session.display_name,
                })

        # Show the Remote Desktop Window
        self.desktop_window = arcane_forms.DesktopWindow(self, self.session)
        self.desktop_window.show()



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.
"""

from typing import Optional, Union

from PyQt6.QtCore import QModelIndex, QSettings, Qt
from PyQt6.QtGui import QShowEvent, QStandardItem, QStandardItemModel
from PyQt6.QtWidgets import (QComboBox, QDialog, QGridLayout, QGroupBox,
                             QHBoxLayout, QLabel, QMainWindow, QMessageBox,
                             QPushButton, QSizePolicy, QSpacerItem, QSpinBox,
                             QTabWidget, QTreeView, QVBoxLayout, QWidget)

import arcane_viewer.arcane as arcane
import arcane_viewer.ui.utilities as utilities

from .options_dialogs import ServerCertificateAddOrEditDialog


class RemoteDesktopOptionsTab(QWidget):
    """ Remote Desktop Options Tab """
    def __init__(self, options_dialog: QDialog, settings: QSettings) -> None:
        super().__init__()

        self.options_dialog = options_dialog

        self.settings = settings

        core_layout = QVBoxLayout()
        self.setLayout(core_layout)

        # Options
        options_layout = QGridLayout()
        options_layout.setContentsMargins(0, 8, 0, 0)
        core_layout.addLayout(options_layout)

        # Clipboard Sharing Mode
        clipboard_sharing_label = QLabel("Clipboard Sharing:")

        self.clipboard_sharing_combobox = QComboBox()
        for clipboard_mode in arcane.ClipboardMode:
            self.clipboard_sharing_combobox.addItem(clipboard_mode.name, userData=clipboard_mode)

        options_layout.addWidget(clipboard_sharing_label, 0, 0)
        options_layout.addWidget(self.clipboard_sharing_combobox, 0, 1)

        # Capture Settings (Fieldset)
        desktop_capture_group = QGroupBox("Capture Settings")
        desktop_capture_group_layout = QGridLayout()
        desktop_capture_group.setLayout(desktop_capture_group_layout)
        core_layout.addWidget(desktop_capture_group)

        desktop_capture_group_layout.setContentsMargins(8, 16, 8, 8)

        # Virtual Desktop Image Quality
        image_quality_label = QLabel("Image Quality:")

        self.image_quality_input = QSpinBox()
        self.image_quality_input.setMinimum(10)
        self.image_quality_input.setMaximum(100)
        self.image_quality_input.setValue(80)

        # Packet Size (Optimization)
        packet_size_label = QLabel("Packet Size:")
        self.packet_size_input = QComboBox()
        for packet_size in arcane.PacketSize:
            self.packet_size_input.addItem(packet_size.display_name, userData=packet_size)

        # Block Size (Optimization)
        block_size_label = QLabel("Block Size:")
        self.block_size_input = QComboBox()
        for block_size in arcane.BlockSize:
            self.block_size_input.addItem(block_size.display_name, userData=block_size)

        # Place Inputs in our Grid Layout
        desktop_capture_group_layout.addWidget(image_quality_label, 0, 0)
        desktop_capture_group_layout.addWidget(self.image_quality_input, 0, 1)

        desktop_capture_group_layout.addWidget(packet_size_label, 1, 0)
        desktop_capture_group_layout.addWidget(self.packet_size_input, 1, 1)

        desktop_capture_group_layout.addWidget(block_size_label, 2, 0)
        desktop_capture_group_layout.addWidget(self.block_size_input, 2, 1)

        core_layout.addItem(QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding))

    def load_settings(self) -> None:
        """ Load remote desktop settings from the settings """
        # Load Options
        self.clipboard_sharing_combobox.setCurrentIndex(
            self.clipboard_sharing_combobox.findData(
                self.settings.value(arcane.SETTINGS_KEY_CLIPBOARD_MODE, arcane.ClipboardMode.Both)
            )
        )

        # Load Capture Options
        self.image_quality_input.setValue(self.settings.value(arcane.SETTINGS_KEY_IMAGE_QUALITY, 80))

        self.packet_size_input.setCurrentIndex(
            self.packet_size_input.findData(
                self.settings.value(arcane.SETTINGS_KEY_PACKET_SIZE, arcane.PacketSize.Size4096)
            )
        )

        self.block_size_input.setCurrentIndex(
            self.block_size_input.findData(
                self.settings.value(arcane.SETTINGS_KEY_BLOCK_SIZE, arcane.BlockSize.Size64)
            )
        )

    def save_settings(self) -> None:
        """ Save remote desktop settings to the settings """
        # Save Options
        self.settings.setValue(arcane.SETTINGS_KEY_CLIPBOARD_MODE, self.clipboard_sharing_combobox.currentData())

        # Save Capture Options
        self.settings.setValue(arcane.SETTINGS_KEY_IMAGE_QUALITY, self.image_quality_input.value())
        self.settings.setValue(arcane.SETTINGS_KEY_PACKET_SIZE, self.packet_size_input.currentData())
        self.settings.setValue(arcane.SETTINGS_KEY_BLOCK_SIZE, self.block_size_input.currentData())


class TrustedCertificateModel(QStandardItemModel):
    """ Trusted Certificate Model (Disables editing of the fingerprint) """
    def flags(self, index:  QModelIndex) -> Qt.ItemFlag:
        if index.column() == 1:  # Fingerprint
            return Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEnabled

        return super().flags(index)


class TrustedCertificatesOptionsTab(QWidget):
    """ Trusted Certificates Options Tab """
    def __init__(self, options_dialog: QDialog, settings: QSettings) -> None:
        super().__init__()

        # I'm using `options_dialog` property instead of parent property `super().__init__(parent)` because it seems
        # that using the parent property is flawed in some cases, until I find a better solution, I will use this
        # method.
        self.options_dialog = options_dialog

        self.settings = settings

        core_layout = QHBoxLayout()
        self.setLayout(core_layout)

        # Setup Certificate List
        self.tree_view = QTreeView()
        core_layout.addWidget(self.tree_view)

        self.model = TrustedCertificateModel()

        self.tree_view.setModel(self.model)
        self.tree_view.setRootIsDecorated(False)

        selection_model = self.tree_view.selectionModel()
        if selection_model is not None:
            selection_model.selectionChanged.connect(self.tree_view_selection_changed)

        self.tree_view.setSelectionBehavior(QTreeView.SelectionBehavior.SelectRows)

        # Setup Action Buttons
        action_buttons_layout = QVBoxLayout()
        core_layout.addLayout(action_buttons_layout)

        self.add_button = QPushButton("Add")
        self.add_button.clicked.connect(lambda: self.add_or_edit_certificate(False))
        action_buttons_layout.addWidget(self.add_button)

        self.edit_button = QPushButton("Edit")
        self.edit_button.setEnabled(False)
        self.edit_button.clicked.connect(lambda: self.add_or_edit_certificate(True))
        action_buttons_layout.addWidget(self.edit_button)

        self.remove_button = QPushButton("Remove")
        self.remove_button.setEnabled(False)
        self.remove_button.clicked.connect(self.remove_button_clicked)
        action_buttons_layout.addWidget(self.remove_button)

        action_buttons_layout.addItem(QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding))

    def add_or_edit_row(self, fingerprint: str, display_name: str, description: str) -> None:
        """ Add a certificate to the list """
        if self.model is None:
            return

        # Edit row if it already exists
        for i in range(self.model.rowCount()):
            col_display_name = self.model.item(i, 0)
            col_fingerprint = self.model.item(i, 1)
            col_description = self.model.item(i, 2)

            if any(column is None for column in [
                col_display_name,
                col_fingerprint,
                col_description
            ]):
                continue

            if col_fingerprint.text() == fingerprint:  # type: ignore[union-attr]
                col_display_name.setText(display_name)  # type: ignore[union-attr]
                col_description.setText(description)  # type: ignore[union-attr]

                return

        # (Or) Add a new row
        self.model.appendRow([
            QStandardItem(display_name),
            QStandardItem(fingerprint),
            QStandardItem(description),
        ])

    def tree_view_selection_changed(self) -> None:
        """ Update the state of the action buttons based on the selection """
        b = self.tree_view.currentIndex().isValid()
        self.edit_button.setEnabled(b)
        self.remove_button.setEnabled(b)

    def remove_button_clicked(self) -> None:
        """ Remove the selected certificate from the list """
        selected_index = self.tree_view.currentIndex()
        if not selected_index.isValid():
            return

        self.model.removeRow(selected_index.row())

    def add_or_edit_certificate(self, edit_selected: bool) -> None:
        """ Add or edit a certificate """
        if self.model is None:
            return

        fingerprint = None
        if edit_selected and self.tree_view.currentIndex().isValid():
            col_fingerprint = self.model.item(self.tree_view.currentIndex().row(), 1)
            if col_fingerprint is None:
                return

            fingerprint = col_fingerprint.text()

        dialog = ServerCertificateAddOrEditDialog(self.options_dialog, self.settings, fingerprint)
        dialog.exec()

        if dialog.result() == QDialog.DialogCode.Accepted:
            self.add_or_edit_row(
                dialog.fingerprint_edit.text(),
                dialog.display_name_edit.text(),
                dialog.description_edit.toPlainText()
            )

    def load_settings(self) -> None:
        """ Load trusted certificates from the settings """
        # First we clear the list
        self.model.clear()

        self.model.setHorizontalHeaderLabels([
            "Display Name",
            "SHA-256 Fingerprint",
            "Description",
        ])

        certificates = self.settings.value(arcane.SETTINGS_KEY_TRUSTED_CERTIFICATES, [])

        for certificate in certificates:
            certificate_information = self.settings.value(
                f"{arcane.SETTINGS_KEY_TRUSTED_CERTIFICATES}.{certificate}", {}
            )

            if not all(k in certificate_information for k in (
                "display_name",
                # Additional not mandatory fields must be added here (if any are added in the future)
            )):
                continue

            description = ""
            if "description" in certificate_information:
                description = certificate_information["description"]

            self.add_or_edit_row(certificate, certificate_information["display_name"], description)

        for i in range(self.model.columnCount()):
            self.tree_view.resizeColumnToContents(i)

    def save_settings(self) -> None:
        """ Save trusted certificates to the settings """
        if self.model is None:
            return

        fingerprints = []
        for i in range(self.model.rowCount()):
            col_display_name = self.model.item(i, 0)
            col_fingerprint = self.model.item(i, 1)
            col_description = self.model.item(i, 2)

            if any(column is None for column in [
                col_display_name,
                col_fingerprint,
                col_description
            ]):
                continue

            fingerprint = col_fingerprint.text().upper().strip()  # type: ignore[union-attr]

            fingerprints.append(fingerprint)

            # Save extra information about the certificate
            self.settings.setValue(
                f"{arcane.SETTINGS_KEY_TRUSTED_CERTIFICATES}.{fingerprint}",
                {
                    "display_name": col_display_name.text(),  # type: ignore[union-attr]
                    "description": col_description.text(),  # type: ignore[union-attr]
                    # Additional extra information fields to be placed here (if any are added in the future)
                }
            )

        # Save the list of trusted certificates (Just the fingerprints)
        self.settings.setValue(arcane.SETTINGS_KEY_TRUSTED_CERTIFICATES, fingerprints)


class OptionsDialog(utilities.QCenteredDialog):
    """ Arcane Options Dialog """
    def __init__(self, parent: Optional[Union[QDialog, QMainWindow]] = None) -> None:
        super().__init__(parent)

        self.settings = QSettings(arcane.APP_ORGANIZATION_NAME, arcane.APP_NAME)

        self.setWindowTitle(f"{arcane.APP_DISPLAY_NAME} :: Options")

        self.setWindowFlags(
            Qt.WindowType.Dialog |
            Qt.WindowType.MSWindowsFixedSizeDialogHint
        )

        core_layout = QVBoxLayout()
        self.setLayout(core_layout)

        core_layout.setContentsMargins(8, 8, 8, 8)

        # Setup Options Tab Control
        self.options_tab_widget = QTabWidget()
        core_layout.addWidget(self.options_tab_widget)

        # General Tab
        self.remote_desktop_tab = RemoteDesktopOptionsTab(self, self.settings)
        self.options_tab_widget.addTab(self.remote_desktop_tab, "Remote Desktop")

        # Trusted Certificates Tab
        self.trusted_certificates_tab = TrustedCertificatesOptionsTab(self, self.settings)
        self.options_tab_widget.addTab(self.trusted_certificates_tab, "Trusted Certificates")

        # Action Buttons
        action_buttons_layout = QHBoxLayout()
        core_layout.addLayout(action_buttons_layout)

        self.reset_button = QPushButton('Reset')
        self.reset_button.setObjectName('danger')
        self.reset_button.clicked.connect(self.reset_settings)
        action_buttons_layout.addWidget(self.reset_button)

        spacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        action_buttons_layout.addItem(spacer)

        self.cancel_button = QPushButton('Cancel')
        self.cancel_button.clicked.connect(lambda: self.reject())
        action_buttons_layout.addWidget(self.cancel_button)

        self.save_button = QPushButton('Save')
        self.save_button.clicked.connect(self.save_settings)
        action_buttons_layout.addWidget(self.save_button)

        self.save_button.setDefault(True)
        self.save_button.setFocus()

        self.adjust_size()

    def load_settings(self) -> None:
        self.remote_desktop_tab.load_settings()
        self.trusted_certificates_tab.load_settings()

    def save_settings(self) -> None:
        self.remote_desktop_tab.save_settings()
        self.trusted_certificates_tab.save_settings()

        self.accept()

    def reset_settings(self) -> None:
        if QMessageBox.question(
                self,
                "Reset Settings",
                "Are you sure you want to reset all settings to their defaults? (This action cannot be undone)",
        ) == QMessageBox.StandardButton.Yes:
            self.settings.clear()

            self.load_settings()

    def adjust_size(self) -> None:
        self.setFixedSize(420, self.sizeHint().height())

    def showEvent(self, event: Optional[QShowEvent]) -> None:
        super().showEvent(event)

        self.load_settings()



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.

    Todo:
        - (0001) LogonUI Support
"""

import logging
import struct
from typing import List  # To support python <= 3.8, we need to use `List`
from typing import Optional

from PyQt6.QtCore import QByteArray, QEventLoop, pyqtSignal, pyqtSlot
from PyQt6.QtGui import QImage

import arcane_viewer.arcane as arcane

from .client_base import ClientBaseThread

logger = logging.getLogger(__name__)


class VirtualDesktopThread(ClientBaseThread):
    """ Thread to handle remote desktop streaming, at quantum level """
    open_cellar_door = pyqtSignal(arcane.Screen)
    request_screen_selection = pyqtSignal(list)
    chunk_received = pyqtSignal(QImage, int, int)

    def __init__(self, session: arcane.Session) -> None:
        super().__init__(session, arcane.WorkerKind.Desktop)

        self.selected_screen: Optional[arcane.Screen] = None
        self.event_loop: Optional[QEventLoop] = None

    """`Destruction is a form of creation. So the fact they burn the money is ironic. They just want to see what happens
     when they tear the world apart. They want to change things.`, Donnie Darko"""
    def client_execute(self) -> None:
        if self.client is None:
            return

        screens_obj = self.client.read_json()
        screens = [arcane.Screen(screen) for screen in screens_obj["List"]]
        logger.info(f"{len(screens)} screen(s) detected")

        if len(screens) == 1:
            self.selected_screen = screens[0]
        else:
            self.display_screen_selection_dialog(screens)

        if self.selected_screen is None:
            return

        logger.info(f"Screen: {self.selected_screen.name} "
                    f"({self.selected_screen.width}x{self.selected_screen.height})")

        self.client.write_json(
            {
                "ScreenName": self.selected_screen.name,
                "ImageCompressionQuality": self.session.option_image_quality,
                "PacketSize": self.session.option_packet_size.value,
                "BlockSize": self.session.option_block_size.value,
                "LogonUI": False,  # TODO: 0001
            }
        )

        """ Open Cellar Door
        `This famous linguist once said, of all the phrases in the English language, of all the endless combinations
        of words in all of history, that 'cellar door' is the most beautiful.`, Karen Pomeroy"""
        self.open_cellar_door.emit(
            self.selected_screen
        )

        packet_max_size = self.session.option_packet_size.value
        while self._running:
            try:
                chunk_size, x, y = struct.unpack('III', self.client.conn.read(12))
            except (Exception, ):
                break

            chunk_bytes = QByteArray()
            bytes_read = 0
            while bytes_read < chunk_size:
                packet_size = min(packet_max_size, chunk_size - bytes_read)
                b = self.client.conn.recv(packet_size)
                if not b:
                    break

                bytes_read += len(b)

                chunk_bytes.append(b)

            chunk = QImage()
            chunk.loadFromData(chunk_bytes)

            self.chunk_received.emit(
                chunk,
                x,
                y,
            )

    def stop(self) -> None:
        super().stop()

        if self.event_loop is not None:
            self.event_loop.quit()

    def display_screen_selection_dialog(self, screens: List[arcane.Screen]) -> None:
        self.event_loop = QEventLoop()

        self.request_screen_selection.emit(screens)

        self.event_loop.exec()

    @pyqtSlot(arcane.Screen)
    def on_screen_selection_dialog_closed(self, screen: arcane.Screen) -> None:
        self.selected_screen = screen

        if self.event_loop is not None:
            self.event_loop.quit()
            self.event_loop = None



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.
"""

import re
from typing import Optional, Union

from PyQt6.QtCore import QSettings
from PyQt6.QtWidgets import (QDialog, QHBoxLayout, QLabel, QLineEdit,
                             QMainWindow, QMessageBox, QPushButton,
                             QSizePolicy, QSpacerItem, QTextEdit, QVBoxLayout)

import arcane_viewer.arcane as arcane
import arcane_viewer.ui.utilities as utilities


class ServerCertificateAddOrEditDialog(utilities.QCenteredDialog):
    def __init__(
            self,
            parent: Optional[Union[QDialog, QMainWindow]],
            settings: QSettings, fingerprint: Optional[str] = None
    ) -> None:
        super().__init__(parent)

        self.settings = settings
        self.fingerprint = fingerprint

        self.setWindowTitle("{} Trusted Server Certificate".format(
            "Add" if self.fingerprint is None else "Edit"
        ))

        core_layout = QVBoxLayout()
        self.setLayout(core_layout)

        # Create our form fields
        self.fingerprint_label = QLabel("Fingerprint:")
        core_layout.addWidget(self.fingerprint_label)

        self.fingerprint_edit = QLineEdit()
        self.fingerprint_edit.setEnabled(self.fingerprint is None)
        self.fingerprint_edit.setText(self.fingerprint)
        core_layout.addWidget(self.fingerprint_edit)

        self.display_name_label = QLabel("Display Name:")
        core_layout.addWidget(self.display_name_label)

        self.display_name_edit = QLineEdit()
        core_layout.addWidget(self.display_name_edit)

        self.description_label = QLabel("Description:")
        core_layout.addWidget(self.description_label)

        self.description_edit = QTextEdit()
        self.description_edit.setMaximumHeight(100)
        core_layout.addWidget(self.description_edit)

        # Load the values if we are editing
        certificate_information = self.settings.value(
            f"{arcane.SETTINGS_KEY_TRUSTED_CERTIFICATES}.{self.fingerprint}", None
        )

        if certificate_information is not None:
            if "display_name" in certificate_information:
                self.display_name_edit.setText(certificate_information["display_name"])

            if "description" in certificate_information:
                self.description_edit.setText(certificate_information["description"])

        core_layout.addSpacing(6)

        # Action Buttons
        action_buttons_layout = QHBoxLayout()
        core_layout.addLayout(action_buttons_layout)

        spacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        action_buttons_layout.addItem(spacer)

        self.cancel_button = QPushButton("Cancel")
        self.cancel_button.clicked.connect(lambda: self.reject())
        action_buttons_layout.addWidget(self.cancel_button)

        button_caption = "Add" if self.fingerprint is None else "Save"
        self.validate_button = QPushButton(button_caption)
        self.validate_button.clicked.connect(self.save_or_update_certificate)
        self.validate_button.setDefault(True)
        action_buttons_layout.addWidget(self.validate_button)

        self.adjust_size()

    def save_or_update_certificate(self) -> None:
        """ Save or update the certificate information """
        if self.fingerprint is None:
            fingerprint = self.fingerprint_edit.text().upper()
            if re.match(r'^[a-fA-F0-9]{40}$', fingerprint) is None:
                QMessageBox.critical(
                    self,
                    "Invalid Certificate Fingerprint",
                    "The fingerprint is invalid, it must be a 40 characters hexadecimal string (SHA-1)."
                )

                self.fingerprint_edit.setFocus()

                return

        self.accept()

    def adjust_size(self) -> None:
        self.setFixedSize(400, self.sizeHint().height())



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.
"""

import os
import sys

ASSETS_IDENTIFIER = "arcane_viewer.assets"

# Asset path getter
if sys.version_info < (3, 9):
    import pkg_resources

    def get_asset_file(asset_name: str) -> str:
        return pkg_resources.resource_filename(ASSETS_IDENTIFIER, asset_name)

else:
    import importlib.resources as resources

    def get_asset_file(asset_name: str) -> str:
        asset_path = resources.files(ASSETS_IDENTIFIER) / asset_name
        return str(asset_path)


# Application Information
APP_VERSION = "1.0.5"
APP_NAME = "Arcane"
APP_ORGANIZATION_NAME = "Phrozen"
APP_DISPLAY_NAME = f"{APP_NAME} {APP_VERSION} (βeta)"

# Remote Desktop Engine Hardcoded Values
VD_WINDOW_ADJUST_RATIO = 90

# Assets absolute paths
DEFAULT_JSON = os.path.join(get_asset_file("default.json"))
APP_ICON = os.path.join(get_asset_file("app_icon.png"))

# Settings Key Names
SETTINGS_KEY_TRUSTED_CERTIFICATES = "trusted_certificates"
SETTINGS_KEY_IMAGE_QUALITY = "image_quality"
SETTINGS_KEY_PACKET_SIZE = "packet_size"
SETTINGS_KEY_BLOCK_SIZE = "block_size"
SETTINGS_KEY_CLIPBOARD_MODE = "clipboard_mode"



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.
"""

from typing import Optional, Union

from PyQt6.QtCore import QSize, Qt
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import (QDialog, QHBoxLayout, QLabel, QMainWindow,
                             QProgressBar, QSizePolicy, QSpacerItem,
                             QVBoxLayout)

import arcane_viewer.arcane as arcane
import arcane_viewer.ui.utilities as utilities


class ConnectingDialog(utilities.QCenteredDialog):
    def __init__(self, parent: Optional[Union[QDialog, QMainWindow]] = None) -> None:
        super().__init__(parent)

        self.setWindowTitle("📡 Connecting...")

        window_flags = (Qt.WindowType.Dialog |
                        Qt.WindowType.FramelessWindowHint)

        self.setWindowFlags(window_flags)

        core_layout = QHBoxLayout()
        self.setLayout(core_layout)

        # Display Icon to the Left
        icon = QLabel(self)
        icon.setPixmap(QIcon(arcane.APP_ICON).pixmap(QSize(96, 96)))
        icon.setAlignment(Qt.AlignmentFlag.AlignCenter)
        core_layout.addWidget(icon)

        # Display Information to the Right
        info_layout = QVBoxLayout()
        core_layout.addLayout(info_layout)

        spacer_top = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        info_layout.addItem(spacer_top)

        label = QLabel("Connecting to the remote server, please wait...")
        info_layout.addWidget(label)

        progress_bar = QProgressBar()
        progress_bar.setRange(0, 0)
        info_layout.addWidget(progress_bar)

        spacer_bottom = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        info_layout.addItem(spacer_bottom)



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.

    Description:
        Client class that handles secure communication with a remote
        server using TLS. Supports password authentication through a
        challenge-based mechanism.

    Todo:
        * (0001) Add support for certificate-based authentication
        * (0002) Display dialog for unknown / first-time seen server
                 certificate
        * (0003) Support for server certificate verification
        * (0004) Implement custom exception for authentication failure,
                 display a message box
"""

import binascii
import hashlib
import json
import logging
import socket
import ssl

import arcane_viewer.arcane as arcane

logger = logging.getLogger(__name__)


class Client:
    """ Client class to handle secure communication with remote server
    Things to note:
        * It is possible to read and write to a socket at the same time, so one thread (for example the main thread) can
        write to the socket while a secondary thread reads from it (at the same time).
    """
    def __init__(self, server_address: str, server_port: int, password: str) -> None:
        self.id = -1

        self.info("Connecting to remote server: `{}:{}`...".format(
            server_address,
            server_port
        ))

        self.client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self.debug("Creating and configure new SSL context...")
        self.context = ssl.create_default_context()
        self.context.check_hostname = False
        self.context.verify_mode = ssl.CERT_NONE

        self.conn = self.context.wrap_socket(
            self.client,
            server_hostname=server_address,
        )

        self.id = self.conn.fileno()

        self.debug("Establishing connection to remote server...")
        self.conn.settimeout(10)

        self.conn.connect((server_address, server_port))

        server_certificate = self.conn.getpeercert(binary_form=True)
        if server_certificate is None:
            raise arcane.ArcaneProtocolException(
                arcane.ArcaneProtocolError.MissingServerCertificate
            )

        self.server_fingerprint = hashlib.sha1(server_certificate).hexdigest().upper()
        self.debug(f"Server certificate fingerprint: `{self.server_fingerprint}`")

        self.conn.settimeout(None)

        self.info("Connected! Authenticating with remote server...")

        self.authenticate(password)

        self.info("Authentication successful")

    def __del__(self) -> None:
        self.close()

    def _log(self, level, message: str) -> None:
        """ Log a message with the client ID as prefix
        I prefer to use this method instead of directly configuring the logger which I find to be less straightforward.
        Of course if we would need to reflect the same principle across different classes, we would not use this
        method. having an id (uuid) prefix is really for debugging purposes and solely for the client class.
        (Client Tracking: ON; OFF) """
        logger.log(level, f"[{self.id}] {message}")

    def debug(self, message: str) -> None:
        self._log(logging.DEBUG, message)

    def info(self, message: str) -> None:
        self._log(logging.INFO, message)

    def read_line(self) -> str:
        data = b""

        while True:
            try:
                b = self.conn.recv(1)
            except (Exception, ):
                break

            if not b:
                break

            data += b
            if b == b'\n':
                break

        return data.decode('utf-8').strip()

    def write_line(self, line: str) -> None:
        try:
            self.conn.write(line.encode('utf-8') + b'\r\n')
        except (Exception, ):
            pass

    def read_json(self) -> dict:
        try:
            return json.loads(self.read_line())
        except json.JSONDecodeError:
            return {}

    def write_json(self, data: dict) -> None:
        self.write_line(json.dumps(data))

    def authenticate(self, password: str) -> None:
        self.debug("Request challenge...")

        challenge = self.read_line()

        self.debug(f"Received challenge: `{challenge}`, attempt to solve it with defined password...")

        challenge_solution_as_bytes = hashlib.pbkdf2_hmac(
            "sha512",
            password.encode("utf-8"),
            challenge.encode("utf-8"),
            1000
        )
        challenge_solution = binascii.hexlify(challenge_solution_as_bytes)\
            .decode("utf-8").upper()

        self.debug(f"Challenge solved: `{challenge_solution}`, sending solution to server...")

        self.write_line(challenge_solution)

        response = self.read_line()
        if response != arcane.ArcaneProtocolCommand.Success.name:
            raise arcane.ArcaneProtocolException(
                arcane.ArcaneProtocolError.AuthenticationFailed
            )

    def close(self) -> None:
        self.info("Closing connection...")
        if self.conn is not None:
            try:
                self.conn.shutdown(socket.SHUT_RDWR)
                self.conn.close()
            except OSError:
                pass

        if self.client is not None:
            try:
                self.client.close()
            except OSError:
                # Should not happen (very unlikely but not null)
                pass



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.
"""

import logging
import ssl
from json.decoder import JSONDecodeError

from PyQt6.QtCore import Qt, pyqtSignal, pyqtSlot

import arcane_viewer.arcane as arcane

from .client_base import ClientBaseThread

logger = logging.getLogger(__name__)


class EventsThread(ClientBaseThread):
    update_mouse_cursor = pyqtSignal(Qt.CursorShape)
    update_clipboard = pyqtSignal(str)

    def __init__(self, session: arcane.Session) -> None:
        super().__init__(session, arcane.WorkerKind.Events)

    def client_execute(self) -> None:
        """ Execute the client thread """
        if self.client is None:
            return

        while self._running:
            try:
                event = self.client.read_json()
            except JSONDecodeError:
                continue
            except (OSError, ssl.SSLError, ssl.SSLEOFError):
                break

            if event is None or "Id" not in event:
                continue

            event_id = event["Id"]

            # Handle Cursor Icon Updates and Reflect it on the Virtual Desktop (Native Cursor)
            if event_id == arcane.InputEvent.MouseCursorUpdated.value and "Cursor" in event:
                cursor_name = event["Cursor"]

                # Default
                cursor = Qt.CursorShape.ArrowCursor

                if cursor_name in (arcane.MouseCursorKind.IDC_SIZEALL.name,
                                   cursor_name == arcane.MouseCursorKind.IDC_SIZE.name):
                    cursor = Qt.CursorShape.SizeAllCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_SIZENESW.name:
                    cursor = Qt.CursorShape.SizeBDiagCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_SIZENS.name:
                    cursor = Qt.CursorShape.SizeVerCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_SIZENWSE.name:
                    cursor = Qt.CursorShape.SizeFDiagCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_SIZEWE.name:
                    cursor = Qt.CursorShape.SizeHorCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_UPARROW.name:
                    cursor = Qt.CursorShape.UpArrowCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_WAIT.name:
                    cursor = Qt.CursorShape.WaitCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_APPSTARTING.name:
                    cursor = Qt.CursorShape.BusyCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_CROSS.name:
                    cursor = Qt.CursorShape.CrossCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_HAND.name:
                    cursor = Qt.CursorShape.PointingHandCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_HELP.name:
                    cursor = Qt.CursorShape.WhatsThisCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_IBEAM.name:
                    cursor = Qt.CursorShape.IBeamCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_ICON.name:
                    pass  # Obsolete
                elif cursor_name == arcane.MouseCursorKind.IDC_NO.name:
                    cursor = Qt.CursorShape.ForbiddenCursor

                self.update_mouse_cursor.emit(cursor)
            # Handle Clipboard Updates
            elif event_id == arcane.InputEvent.ClipboardUpdated.value and "Text" in event:
                if self.session.clipboard_mode in {
                    arcane.ClipboardMode.Disabled, arcane.ClipboardMode.Send
                }:
                    continue

                self.update_clipboard.emit(event["Text"])

    @pyqtSlot(int, int, arcane.MouseState, arcane.MouseButton)
    def send_mouse_event(self, x: int, y: int, state: arcane.MouseState, button: arcane.MouseButton) -> None:
        """ Send mouse event to the server """
        if self.session.presentation:
            return

        if self.client is not None and self._connected:
            self.client.write_json(
                {
                    "Id": arcane.OutputEvent.MouseClickMove.name,
                    "X": x,
                    "Y": y,
                    "Button": button.name,
                    "Type": state.name,
                }
            )

    @pyqtSlot(str)
    def send_key_event(self, keys: str) -> None:
        """ Send keyboard event to the server """
        if self.session.presentation:
            return

        if self.client is not None and self._connected:
            self.client.write_json(
                {
                    "Id": arcane.OutputEvent.Keyboard.name,
                    "Keys": keys,
                }
            )

    @pyqtSlot(int)
    def send_mouse_wheel_event(self, delta: int) -> None:
        """ Send mouse wheel event to the server """
        if self.session.presentation:
            return

        if self.client is not None and self._connected:
            self.client.write_json(
                {
                    "Id": arcane.OutputEvent.MouseWheel.name,
                    "Delta": delta,
                }
            )

    @pyqtSlot(str)
    def send_clipboard_text(self, text: str) -> None:
        """ Send clipboard text to the server """
        if self.session.clipboard_mode in {
            arcane.ClipboardMode.Disabled, arcane.ClipboardMode.Receive
        }:
            return

        if self.client is not None and self._connected:
            self.client.write_json(
                {
                    "Id": arcane.OutputEvent.ClipboardUpdated.name,
                    "Text": text,
                }
            )



"""
    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    More information about the LICENSE on the LICENSE file in the root directory of the project.
"""

from typing import List, Optional, Union

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (QComboBox, QDialog, QHBoxLayout, QLabel,
                             QMainWindow, QPushButton, QVBoxLayout)

import arcane_viewer.arcane as arcane
import arcane_viewer.ui.utilities as utilities


class ScreenSelectionDialog(utilities.QCenteredDialog):
    """ Screen Selection Dialog """
    def __init__(self, parent: Optional[Union[QDialog, QMainWindow]], screens: List[arcane.Screen]) -> None:
        super().__init__(parent)

        self.screens = screens

        self.setWindowTitle(f"Screen Selection ({len(self.screens)} Available)")

        self.setWindowFlags(
            Qt.WindowType.Dialog |
            Qt.WindowType.CustomizeWindowHint |
            Qt.WindowType.MSWindowsFixedSizeDialogHint
        )

        core_layout = QVBoxLayout()
        self.setLayout(core_layout)

        # Screen Selection Combo
        screen_selection_label = QLabel("Select Screen:")
        core_layout.addWidget(screen_selection_label)

        self.screen_selection_combobox = QComboBox()

        # screen_selection_combobox.addItems([screen.name for screen in screens])
        for screen in self.screens:
            self.screen_selection_combobox.addItem(
                screen.get_display_name(),
                userData=screen.name
            )

        core_layout.addWidget(self.screen_selection_combobox)

        core_layout.addSpacing(8)

        # Action Buttons
        action_buttons_layout = QHBoxLayout()
        core_layout.addLayout(action_buttons_layout)

        self.cancel_button = QPushButton('Cancel')
        self.cancel_button.clicked.connect(lambda: self.reject())
        action_buttons_layout.addWidget(self.cancel_button)

        self.select_button = QPushButton('Select')
        self.select_button.clicked.connect(lambda: self.accept())
        action_buttons_layout.addWidget(self.select_button)

        self.select_button.setDefault(True)
        self.select_button.setFocus()

        self.setFixedSize(290, self.sizeHint().height())

    def get_selected_screen(self) -> arcane.Screen:
        """ Get the user-choice selected screen """
        return [screen for screen in self.screens if screen.name == self.screen_selection_combobox.currentData()][0]
